"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getCodecMetaLookupFromInput = exports.makePgTypeCodecMeta = void 0;
require("graphile-build");
const pg_1 = require("@dataplan/pg");
/**
 * Creates an empty meta object for the given codec.
 */
function makePgTypeCodecMeta(_codec) {
    return {
        typeNameBySituation: Object.create(null),
    };
}
exports.makePgTypeCodecMeta = makePgTypeCodecMeta;
/**
 * Given the input object, this function walks through all the pgSources and
 * all their codecs and relations and extracts the full set of reachable
 * codecs.
 *
 * Memoized for performance, using a WeakMap.
 */
function getCodecMetaLookupFromInput(input) {
    const metaLookup = new Map();
    const seenSources = new Set();
    if (input.pgCodecs) {
        for (const codec of input.pgCodecs) {
            walkCodec(codec, metaLookup);
        }
    }
    for (const source of input.pgSources) {
        walkSource(resolveSource(source), metaLookup, seenSources);
    }
    return metaLookup;
}
exports.getCodecMetaLookupFromInput = getCodecMetaLookupFromInput;
/**
 * Walks the given source's codecs/relations and pushes all discovered codecs
 * into `metaLookup`
 *
 * @internal
 */
function walkSource(source, metaLookup, seenSources) {
    if (seenSources.has(source)) {
        return;
    }
    seenSources.add(source);
    if (!metaLookup.has(source.codec)) {
        walkCodec(source.codec, metaLookup);
    }
    const relations = source.getRelations();
    if (relations) {
        for (const relationshipName in relations) {
            walkSource(resolveSource(relations[relationshipName].source), metaLookup, seenSources);
        }
    }
}
/**
 * Adds the given codec to `metaLookup` and also walks the related codecs (for
 * columns, and inner-codecs).
 *
 * @internal
 */
function walkCodec(codec, metaLookup) {
    if (metaLookup.has(codec)) {
        return;
    }
    metaLookup.set(codec, makePgTypeCodecMeta(codec));
    if (codec.columns) {
        for (const columnName in codec.columns) {
            walkCodec(codec.columns[columnName].codec, metaLookup);
        }
    }
    if (codec.arrayOfCodec) {
        walkCodec(codec.arrayOfCodec, metaLookup);
    }
    if (codec.domainOfCodec) {
        walkCodec(codec.domainOfCodec, metaLookup);
    }
    if (codec.rangeOfCodec) {
        walkCodec(codec.rangeOfCodec, metaLookup);
    }
}
/**
 * Where a source might be a PgSource or a PgSourceBuilder, this ensures we
 * return a PgSource.
 */
function resolveSource(source) {
    return source instanceof pg_1.PgSourceBuilder ? source.get() : source;
}
//# sourceMappingURL=inputUtils.js.map