"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PgColumnsPlugin = void 0;
require("graphile-build");
require("./PgTablesPlugin.js");
require("../interfaces.js");
require("graphile-config");
const pg_1 = require("@dataplan/pg");
const graphile_export_1 = require("graphile-export");
const behavior_js_1 = require("../behavior.js");
const version_js_1 = require("../version.js");
function unwrapCodec(codec) {
    if (codec.arrayOfCodec) {
        return unwrapCodec(codec.arrayOfCodec);
    }
    return codec;
}
const getSource = (0, graphile_export_1.EXPORTABLE)((PgSource) => (baseCodec, pgSources, $record) => {
    const executor = $record.source.executor;
    const source = pgSources.find((potentialSource) => potentialSource.executor === executor) ?? PgSource.fromCodec(executor, baseCodec);
    return source;
}, [pg_1.PgSource]);
function processColumn(fields, build, context, columnName, overrideName) {
    const { extend, graphql: { getNullableType, GraphQLNonNull, GraphQLList }, inflection, getGraphQLTypeByPgCodec, } = build;
    const { scope: { pgCodec }, } = context;
    if (!pgCodec) {
        return;
    }
    const isInterface = context.type === "GraphQLInterfaceType";
    const column = pgCodec.columns[columnName];
    const behavior = (0, behavior_js_1.getBehavior)([pgCodec.extensions, column.extensions]);
    if (!build.behavior.matches(behavior, "attribute:select", "select")) {
        // Don't allow selecting this column.
        return;
    }
    const columnFieldName = overrideName ??
        inflection.column({
            columnName,
            codec: pgCodec,
        });
    const baseCodec = unwrapCodec(column.codec);
    const baseType = getGraphQLTypeByPgCodec(baseCodec, "output");
    const arrayOrNotType = column.codec.arrayOfCodec
        ? new GraphQLList(baseType)
        : baseType;
    if (!arrayOrNotType) {
        console.warn(`Couldn't find a 'output' variant for PgTypeCodec ${pgCodec.name}'s '${columnName}' column (${column.codec.name}; array=${!!column.codec
            .arrayOfCodec}, domain=${!!column.codec.domainOfCodec}, enum=${!!column.codec.values})`);
        return;
    }
    const type = column.notNull
        ? new GraphQLNonNull(getNullableType(arrayOrNotType))
        : arrayOrNotType;
    if (!type) {
        // Could not determine the type, skip this field
        console.warn(`Could not determine the type for column '${columnName}' of ${pgCodec.name}`);
        return;
    }
    const fieldSpec = {
        description: column.description,
        type: type,
    };
    if (!isInterface) {
        const makePlan = () => {
            // See if there's a source to pull record types from (e.g. for relations/etc)
            if (!baseCodec.columns) {
                // Simply get the value
                return (0, graphile_export_1.EXPORTABLE)((columnName) => ($record) => {
                    return $record.get(columnName);
                }, [columnName]);
            }
            else {
                const pgSources = build.input.pgSources.filter((potentialSource) => potentialSource.codec === baseCodec && !potentialSource.parameters);
                // TODO: this is pretty horrible in the export; we should fix that.
                if (!column.codec.arrayOfCodec) {
                    const notNull = column.notNull || column.codec.notNull;
                    // Single record from source
                    /*
                     * TODO: if we refactor `PgSelectSingleStep` we can probably
                     * optimise this to do inline selection and still join against
                     * the base table using e.g. `(table.column).attribute =
                     * joined_thing.column`
                     */
                    return (0, graphile_export_1.EXPORTABLE)((baseCodec, columnName, getSource, notNull, pgSelectSingleFromRecord, pgSources) => ($record) => {
                        const $plan = $record.get(columnName);
                        const $select = pgSelectSingleFromRecord(getSource(baseCodec, pgSources, $record), $plan);
                        if (notNull) {
                            $select.coalesceToEmptyObject();
                        }
                        $select.getClassStep().setTrusted();
                        return $select;
                    }, [
                        baseCodec,
                        columnName,
                        getSource,
                        notNull,
                        pg_1.pgSelectSingleFromRecord,
                        pgSources,
                    ]);
                }
                else {
                    // Many records from source
                    /*
                     * TODO: if we refactor `PgSelectSingleStep` we can probably
                     * optimise this to do inline selection and still join against
                     * the base table using e.g. `(table.column).attribute =
                     * joined_thing.column`
                     */
                    return (0, graphile_export_1.EXPORTABLE)((baseCodec, columnName, getSource, pgSelectFromRecords, pgSources) => ($record) => {
                        const $val = $record.get(columnName);
                        const $select = pgSelectFromRecords(getSource(baseCodec, pgSources, $record), $val);
                        $select.setTrusted();
                        return $select;
                    }, [baseCodec, columnName, getSource, pg_1.pgSelectFromRecords, pgSources]);
                }
            }
        };
        fieldSpec.plan = makePlan();
    }
    fields = extend(fields, {
        [columnFieldName]: context.fieldWithHooks({
            fieldName: columnFieldName,
            pgColumn: column,
        }, fieldSpec),
    }, `Adding '${columnName}' column field to PgTypeCodec '${pgCodec.name}'`);
}
exports.PgColumnsPlugin = {
    name: "PgColumnsPlugin",
    description: "Adds PostgreSQL columns (attributes) to the relevant GraphQL object/input object types",
    version: version_js_1.version,
    // TODO: Requires PgTablesPlugin
    inflection: {
        add: {
            _columnName(options, { columnName, codec }) {
                const column = codec.columns[columnName];
                return this.coerceToGraphQLName(column.extensions?.tags?.name || columnName);
            },
            _joinColumnNames(options, codec, names) {
                return names
                    .map((columnName) => {
                    return this.column({ columnName, codec });
                })
                    .join("-and-");
            },
            column(options, details) {
                const columnFieldName = this.camelCase(this._columnName(details));
                // Avoid conflict with 'id' field used for Relay.
                return columnFieldName === "id" && !details.codec.isAnonymous
                    ? "rowId"
                    : columnFieldName;
            },
        },
    },
    schema: {
        hooks: {
            GraphQLInterfaceType_fields(fields, build, context) {
                const { scope: { pgCodec, pgPolymorphism }, } = context;
                if (!pgPolymorphism || !pgCodec?.columns) {
                    return fields;
                }
                for (const columnName in pgCodec.columns) {
                    switch (pgPolymorphism.mode) {
                        case "single": {
                            if (!pgPolymorphism.commonColumns.includes(columnName)) {
                                continue;
                            }
                            break;
                        }
                        case "relational": {
                            break;
                        }
                        case "union": {
                            break;
                        }
                        default: {
                            const never = pgPolymorphism;
                            throw new Error(`Unhandled polymorphism mode ${never.mode}}`);
                        }
                    }
                    processColumn(fields, build, context, columnName);
                }
                return fields;
            },
            GraphQLObjectType_fields(fields, build, context) {
                const { scope: { pgCodec, isPgTableType, pgPolymorphism, pgPolymorphicSingleTableType, }, } = context;
                if (!isPgTableType || !pgCodec?.columns) {
                    return fields;
                }
                for (const columnName in pgCodec.columns) {
                    let overrideName = undefined;
                    if (pgPolymorphism) {
                        switch (pgPolymorphism.mode) {
                            case "single": {
                                const match = pgPolymorphicSingleTableType?.columns.find((c) => c.column === columnName);
                                if (!pgPolymorphism.commonColumns.includes(columnName) &&
                                    !match) {
                                    continue;
                                }
                                if (match?.rename) {
                                    overrideName = match.rename;
                                }
                                break;
                            }
                            case "relational": {
                                break;
                            }
                            case "union": {
                                break;
                            }
                            default: {
                                const never = pgPolymorphism;
                                throw new Error(`Unhandled polymorphism mode ${never.mode}}`);
                            }
                        }
                    }
                    processColumn(fields, build, context, columnName, overrideName);
                }
                return fields;
            },
            GraphQLInputObjectType_fields(fields, build, context) {
                const { extend, inflection } = build;
                const { scope: { isPgRowType, isPgCompoundType, isPgPatch, isPgBaseInput, pgCodec, }, fieldWithHooks, } = context;
                if (!(isPgRowType || isPgCompoundType) ||
                    !pgCodec ||
                    !pgCodec.columns ||
                    pgCodec.isAnonymous) {
                    return fields;
                }
                return Object.entries(pgCodec.columns).reduce((memo, [columnName, column]) => build.recoverable(memo, () => {
                    const behavior = (0, behavior_js_1.getBehavior)([
                        pgCodec.extensions,
                        column.extensions,
                    ]);
                    const action = isPgBaseInput
                        ? "base"
                        : isPgPatch
                            ? "update"
                            : "insert";
                    const fieldBehaviorScope = `attribute:${action}`;
                    if (!build.behavior.matches(behavior, fieldBehaviorScope, action)) {
                        return memo;
                    }
                    const fieldName = inflection.column({
                        columnName,
                        codec: pgCodec,
                    });
                    if (memo[fieldName]) {
                        throw new Error(`Two columns produce the same GraphQL field name '${fieldName}' on input PgTypeCodec '${pgCodec.name}'; one of them is '${columnName}'`);
                    }
                    const columnType = build.getGraphQLTypeByPgCodec(column.codec, "input");
                    if (!columnType) {
                        return memo;
                    }
                    return extend(memo, {
                        [fieldName]: fieldWithHooks({
                            fieldName,
                            fieldBehaviorScope,
                            pgCodec,
                            pgColumn: column,
                        }, {
                            description: column.description,
                            type: build.nullableIf(isPgBaseInput ||
                                isPgPatch ||
                                (!column.notNull &&
                                    !column.extensions?.tags?.notNull) ||
                                column.hasDefault ||
                                Boolean(column.extensions?.tags?.hasDefault), columnType),
                            applyPlan: (0, graphile_export_1.EXPORTABLE)((columnName) => function plan($insert, val) {
                                $insert.set(columnName, val.get());
                            }, [columnName]),
                        }),
                    }, `Adding input object field for ${pgCodec.name}.`);
                }), fields);
            },
        },
    },
};
//# sourceMappingURL=PgColumnsPlugin.js.map