"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PgConditionArgumentPlugin = void 0;
require("./PgTablesPlugin.js");
require("graphile-config");
const graphile_export_1 = require("graphile-export");
const behavior_js_1 = require("../behavior.js");
const version_js_1 = require("../version.js");
exports.PgConditionArgumentPlugin = {
    name: "PgConditionArgumentPlugin",
    description: "Adds the 'condition' argument to connections and lists",
    version: version_js_1.version,
    inflection: {
        add: {
            conditionType(options, typeName) {
                return this.upperCamelCase(`${typeName}-condition`);
            },
        },
    },
    schema: {
        hooks: {
            init(_, build) {
                const { inflection, sql } = build;
                for (const codec of build.pgCodecMetaLookup.keys()) {
                    build.recoverable(null, () => {
                        // Ignore scalar codecs
                        if (!codec.columns || codec.isAnonymous) {
                            return;
                        }
                        const behavior = (0, behavior_js_1.getBehavior)(codec.extensions);
                        // TODO: do we want this filter here? E.g. we might want to enable a bulk delete mutation without allowing any selects?
                        if (!build.behavior.matches(behavior, "select", "select")) {
                            return;
                        }
                        const tableTypeName = inflection.tableType(codec);
                        const conditionName = inflection.conditionType(tableTypeName);
                        /* const TableConditionType = */
                        build.registerInputObjectType(conditionName, {
                            isPgCondition: true,
                            pgCodec: codec,
                        }, () => ({
                            description: build.wrapDescription(`A condition to be used against \`${tableTypeName}\` object types. All fields are tested for equality and combined with a logical ‘and.’`, "type"),
                            fields: (context) => {
                                const { fieldWithHooks } = context;
                                const columns = codec.columns;
                                // TODO: move this to a separate plugin
                                return Object.entries(columns).reduce((memo, [columnName, column]) => {
                                    const behavior = (0, behavior_js_1.getBehavior)([
                                        codec.extensions,
                                        column.extensions,
                                    ]);
                                    if (!build.behavior.matches(behavior, "attribute:filterBy", "filterBy")) {
                                        return memo;
                                    }
                                    // TODO: add `attribute:filterBy:array`/`:range` ?
                                    const fieldName = inflection.column({
                                        columnName,
                                        codec,
                                    });
                                    const type = build.getGraphQLTypeByPgCodec(column.codec, "input");
                                    if (!type) {
                                        return memo;
                                    }
                                    memo = build.extend(memo, {
                                        [fieldName]: fieldWithHooks({
                                            fieldName,
                                            fieldBehaviorScope: "attribute:filterBy",
                                            isPgConnectionConditionInputField: true,
                                        }, {
                                            description: build.wrapDescription(`Checks for equality with the object’s \`${fieldName}\` field.`, "field"),
                                            type: type,
                                            applyPlan: (0, graphile_export_1.EXPORTABLE)((column, columnName, sql) => function plan($condition, val) {
                                                if (val.getRaw().evalIs(null)) {
                                                    $condition.where({
                                                        type: "attribute",
                                                        attribute: columnName,
                                                        callback: (expression) => sql `${expression} is null`,
                                                    });
                                                }
                                                else {
                                                    $condition.where({
                                                        type: "attribute",
                                                        attribute: columnName,
                                                        callback: (expression) => sql `${expression} = ${$condition.placeholder(val.get(), column.codec)}`,
                                                    });
                                                }
                                            }, [column, columnName, sql]),
                                        }),
                                    }, `Adding condition argument for ${codec.name}' ${columnName} column`);
                                    return memo;
                                }, Object.create(null));
                            },
                        }), `Adding condition type for ${codec.name}.`);
                    });
                }
                return _;
            },
            GraphQLObjectType_fields_field_args(args, build, context) {
                const { scope, Self } = context;
                const { fieldName, fieldBehaviorScope, isPgFieldConnection, isPgFieldSimpleCollection, pgSource, pgFieldCodec, } = scope;
                const shouldAddCondition = isPgFieldConnection || isPgFieldSimpleCollection;
                const codec = pgFieldCodec ?? pgSource?.codec;
                const isSuitableSource = pgSource && pgSource.codec.columns && !pgSource.isUnique;
                const isSuitableCodec = codec &&
                    (isSuitableSource ||
                        (!pgSource && codec?.polymorphism?.mode === "union")) &&
                    codec.columns;
                if (!shouldAddCondition || !isSuitableCodec) {
                    return args;
                }
                const behavior = (0, behavior_js_1.getBehavior)([
                    scope,
                    codec?.extensions,
                    pgSource?.extensions,
                ]);
                if (!build.behavior.matches(behavior, fieldBehaviorScope ? `${fieldBehaviorScope}:filter` : `filter`, pgSource?.parameters ? "" : "filter")) {
                    return args;
                }
                const tableTypeName = build.inflection.tableType(codec);
                const tableConditionTypeName = build.inflection.conditionType(tableTypeName);
                const tableConditionType = build.getTypeByName(tableConditionTypeName);
                if (!tableConditionType) {
                    return args;
                }
                return build.extend(args, {
                    condition: {
                        description: build.wrapDescription("A condition to be used in determining which values should be returned by the collection.", "arg"),
                        type: tableConditionType,
                        applyPlan: isPgFieldConnection
                            ? (_condition, $connection) => {
                                const $select = $connection.getSubplan();
                                return $select.wherePlan();
                            }
                            : (_condition, $select) => {
                                return $select.wherePlan();
                            },
                    },
                }, `Adding condition to connection field '${fieldName}' of '${Self.name}'`);
            },
        },
    },
};
//# sourceMappingURL=PgConditionArgumentPlugin.js.map