"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PgIntrospectionPlugin = void 0;
require("graphile-build");
const pg_1 = require("@dataplan/pg");
const grafast_1 = require("grafast");
const graphile_export_1 = require("graphile-export");
const pg_introspection_1 = require("pg-introspection");
const pgConfigs_js_1 = require("../pgConfigs.js");
const version_js_1 = require("../version.js");
const watchFixtures_js_1 = require("../watchFixtures.js");
async function getDb(info, databaseName) {
    const introspections = await info.helpers.pgIntrospection.getIntrospection();
    const relevant = introspections.find((intro) => intro.pgConfig.name === databaseName);
    if (!relevant) {
        throw new Error(`Could not find database '${databaseName}'`);
    }
    return relevant;
}
function makeGetEntity(loc) {
    return async (info, databaseName, id) => {
        const relevant = await getDb(info, databaseName);
        const list = relevant.introspection[loc];
        if (!list) {
            throw new Error(`Could not find database '${databaseName}''s introspection results for '${loc}'`);
        }
        return list.find((entity) => "_id" in entity
            ? entity._id === id
            : "indexrelid" in entity
                ? entity.indexrelid
                : false);
    };
}
function makeGetEntities(loc) {
    return async (info, databaseName) => {
        const relevant = await getDb(info, databaseName);
        const list = relevant.introspection[loc];
        if (!list) {
            throw new Error(`Could not find database '${databaseName}''s introspection results for '${loc}'`);
        }
        return list;
    };
}
exports.PgIntrospectionPlugin = {
    name: "PgIntrospectionPlugin",
    description: "Introspects PostgreSQL databases and makes the results available to other plugins",
    version: version_js_1.version,
    // TODO: refactor TypeScript so this isn't necessary; maybe via `makePluginGatherConfig`?
    gather: {
        namespace: "pgIntrospection",
        initialCache: () => ({
            introspectionResultsPromise: null,
        }),
        initialState: () => ({
            introspectionResultsPromise: null,
            executors: Object.create(null),
        }),
        helpers: {
            getExecutorForDatabase(info, databaseName) {
                if (info.state.executors[databaseName]) {
                    return info.state.executors[databaseName];
                }
                const pgConfig = info.resolvedPreset.pgConfigs?.find((db) => db.name === databaseName);
                if (!pgConfig) {
                    throw new Error(`Database '${databaseName}' not found`);
                }
                const { pgSettingsKey, withPgClientKey } = pgConfig;
                const executor = (0, graphile_export_1.EXPORTABLE)((PgExecutor, constant, context, databaseName, object, pgSettingsKey, withPgClientKey) => new PgExecutor({
                    name: databaseName,
                    context: () => {
                        const ctx = context();
                        return object({
                            pgSettings: pgSettingsKey != null
                                ? ctx.get(pgSettingsKey)
                                : constant(null),
                            withPgClient: ctx.get(withPgClientKey),
                        });
                    },
                }), [
                    pg_1.PgExecutor,
                    grafast_1.constant,
                    grafast_1.context,
                    databaseName,
                    grafast_1.object,
                    pgSettingsKey,
                    withPgClientKey,
                ]);
                info.state.executors[databaseName] = executor;
                return executor;
            },
            getNamespace: makeGetEntity("namespaces"),
            getClasses: makeGetEntities("classes"),
            getClass: makeGetEntity("classes"),
            getConstraint: makeGetEntity("constraints"),
            getProc: makeGetEntity("procs"),
            getRoles: makeGetEntity("roles"),
            getType: makeGetEntity("types"),
            getEnum: makeGetEntity("enums"),
            getExtension: makeGetEntity("extensions"),
            getIndex: makeGetEntity("indexes"),
            getLanguage: makeGetEntity("languages"),
            // TODO: we need getters for these
            // getAuthMembers: makeGetEntity("authMembers"),
            // getRange: makeGetEntity("ranges"),
            // getDepend: makeGetEntity("depends"),
            // getDescription: makeGetEntity("descriptions"),
            //
            async getAttribute(info, databaseName, classId, attributeNumber) {
                // const pgClass = this.getClass(info, databaseName, classId);
                const attributes = await this.getAttributesForClass(info, databaseName, classId);
                return attributes.find((attr) => attr.attnum === attributeNumber);
            },
            async getAttributesForClass(info, databaseName, classId) {
                // const pgClass = this.getClass(info, databaseName, classId);
                const relevant = await getDb(info, databaseName);
                const list = relevant.introspection.attributes;
                // TODO: cache
                return list.filter((entity) => entity.attrelid === classId);
            },
            async getConstraintsForClass(info, databaseName, classId) {
                // const pgClass = this.getClass(info, databaseName, classId);
                const relevant = await getDb(info, databaseName);
                const list = relevant.introspection.constraints;
                // TODO: cache
                return list.filter((entity) => entity.conrelid === classId);
            },
            async getForeignConstraintsForClass(info, databaseName, classId) {
                // const pgClass = this.getClass(info, databaseName, classId);
                const relevant = await getDb(info, databaseName);
                const list = relevant.introspection.constraints;
                // TODO: cache
                return list.filter((entity) => entity.confrelid === classId);
            },
            async getInheritedForClass(info, databaseName, classId) {
                // const pgClass = this.getClass(info, databaseName, classId);
                const relevant = await getDb(info, databaseName);
                const list = relevant.introspection.inherits;
                // TODO: cache
                return list.filter((entity) => entity.inhrelid === classId);
            },
            async getNamespaceByName(info, databaseName, name) {
                const relevant = await getDb(info, databaseName);
                const list = relevant.introspection.namespaces;
                return list.find((nsp) => nsp.nspname === name);
            },
            async getClassByName(info, databaseName, schemaName, tableName) {
                const relevant = await getDb(info, databaseName);
                const list = relevant.introspection.classes;
                return list.find((rel) => rel.getNamespace().nspname === schemaName &&
                    rel.relname === tableName);
            },
            // TODO: we should maybe use pg_type.typelem and look up by ID directy
            // instead of having this function
            async getTypeByArray(info, databaseName, arrayId) {
                const relevant = await getDb(info, databaseName);
                const list = relevant.introspection.types;
                return list.find((type) => type.typarray === arrayId);
            },
            async getEnumsForType(info, databaseName, typeId) {
                const relevant = await getDb(info, databaseName);
                const list = relevant.introspection.enums;
                // TODO: cache
                return list
                    .filter((entity) => entity.enumtypid === typeId)
                    .sort((a, z) => a.enumsortorder - z.enumsortorder);
            },
            async getRangeByType(info, databaseName, typeId) {
                const relevant = await getDb(info, databaseName);
                const list = relevant.introspection.ranges;
                // TODO: cache
                return list.find((entity) => entity.rngtypid === typeId);
            },
            async getExtensionByName(info, databaseName, extensionName) {
                const relevant = await getDb(info, databaseName);
                const list = relevant.introspection.extensions;
                // TODO: cache
                return list.find((entity) => entity.extname === extensionName);
            },
            getIntrospection(info) {
                // IMPORTANT: introspection shouldn't change within a single run (even
                // if the cache does), thus we add it to state.
                if (info.state.introspectionResultsPromise) {
                    return info.state.introspectionResultsPromise;
                }
                info.state.introspectionResultsPromise = (async () => {
                    if (info.cache.introspectionResultsPromise) {
                        return info.cache.introspectionResultsPromise;
                    }
                    if (!info.resolvedPreset.pgConfigs) {
                        return [];
                    }
                    const seenNames = new Map();
                    const seenPgSettingsKeys = new Map();
                    const seenWithPgClientKeys = new Map();
                    // Resolve the promise ASAP so dependents can `getIntrospection()` and then `getClass` or whatever from the result.
                    const introspectionPromise = Promise.all(info.resolvedPreset.pgConfigs.map(async (pgConfig, i) => {
                        // Validate there's no conflicts between pgConfigs
                        const { name, pgSettingsKey, withPgClientKey } = pgConfig;
                        if (!name) {
                            throw new Error(`pgConfigs[${i}] has no name`);
                        }
                        if (!withPgClientKey) {
                            throw new Error(`pgConfigs[${i}] has no withPgClientKey`);
                        }
                        {
                            const existingIndex = seenNames.get(name);
                            if (existingIndex != null) {
                                throw new Error(`pgConfigs[${i}] has the same name as pgConfigs[${existingIndex}] (${JSON.stringify(name)})`);
                            }
                            seenNames.set(name, i);
                        }
                        {
                            const existingIndex = seenWithPgClientKeys.get(withPgClientKey);
                            if (existingIndex != null) {
                                throw new Error(`pgConfigs[${i}] has the same withPgClientKey as pgConfigs[${existingIndex}] (${JSON.stringify(withPgClientKey)})`);
                            }
                            seenWithPgClientKeys.set(withPgClientKey, i);
                        }
                        if (pgSettingsKey) {
                            const existingIndex = seenPgSettingsKeys.get(pgSettingsKey);
                            if (existingIndex != null) {
                                throw new Error(`pgConfigs[${i}] has the same pgSettingsKey as pgConfigs[${existingIndex}] (${JSON.stringify(pgSettingsKey)})`);
                            }
                            seenPgSettingsKeys.set(pgSettingsKey, i);
                        }
                        // Do the introspection
                        const introspectionQuery = (0, pg_introspection_1.makeIntrospectionQuery)();
                        const { rows: [row], } = await (0, pgConfigs_js_1.withPgClientFromPgConfig)(pgConfig, pgConfig.pgSettingsForIntrospection ?? null, (client) => client.query({
                            text: introspectionQuery,
                        }));
                        if (!row) {
                            throw new Error("Introspection failed");
                        }
                        const introspection = (0, pg_introspection_1.parseIntrospectionResults)(row.introspection);
                        return { pgConfig, introspection };
                    }));
                    info.cache.introspectionResultsPromise = introspectionPromise;
                    // Don't cache errors
                    introspectionPromise.then(null, () => {
                        info.cache.introspectionResultsPromise = null;
                    });
                    return introspectionPromise;
                })();
                return info.state.introspectionResultsPromise;
            },
            async getDatabase(info, databaseName) {
                const all = await info.helpers.pgIntrospection.getIntrospection();
                const match = all.find((n) => n.pgConfig.name === databaseName);
                if (!match) {
                    throw new Error(`Could not find results for database '${databaseName}'`);
                }
                return match;
            },
        },
        async watch(info, callback) {
            const unlistens = [];
            for (const pgConfig of info.resolvedPreset.pgConfigs ?? []) {
                if (!pgConfig.pgSubscriber) {
                    console.warn(`pgConfig '${pgConfig.name}' does not have a pgSubscriber, and thus cannot be used for watch mode`);
                    continue;
                }
                // install the watch fixtures
                if (info.options.installWatchFixtures ?? true) {
                    try {
                        await (0, pgConfigs_js_1.withSuperuserPgClientFromPgConfig)(pgConfig, null, (client) => client.query({ text: watchFixtures_js_1.watchFixtures }));
                    }
                    catch (e) {
                        console.warn(`Failed to install watch fixtures into '${pgConfig.name}'.\nInstalling watch fixtures requires superuser privileges; have you correctly configured a 'superuserConnectionString'?\nYou may also opt to configure 'installWatchFixtures: false' and install them yourself.\n\nPostgres says: ${e}`);
                    }
                }
                try {
                    const eventStream = await pgConfig.pgSubscriber.subscribe("postgraphile_watch");
                    const $$stop = Symbol("stop");
                    const abort = (0, grafast_1.defer)();
                    unlistens.push(() => abort.resolve($$stop));
                    const waitNext = () => {
                        const next = Promise.race([abort, eventStream.next()]);
                        next.then((event) => {
                            if (event === $$stop) {
                                // Terminate the stream
                                if (eventStream.return) {
                                    eventStream.return();
                                }
                                else if (eventStream.throw) {
                                    eventStream.throw(new Error("Please stop streaming events now."));
                                }
                            }
                            else {
                                try {
                                    // Delete the introspection results
                                    info.cache.introspectionResultsPromise = null;
                                    // Trigger re-gather
                                    callback();
                                }
                                finally {
                                    // Wait for the next event
                                    waitNext();
                                }
                            }
                        }, (e) => {
                            console.error(`Unexpected error occurred while watching pgConfig '${pgConfig.name}' for schema changes.`, e);
                            abort.resolve($$stop);
                        });
                    };
                    waitNext();
                }
                catch (e) {
                    console.warn(`Failed to watch '${pgConfig.name}': ${e}`);
                }
            }
            return () => {
                for (const cb of unlistens) {
                    try {
                        cb();
                    }
                    catch {
                        /*nom nom nom*/
                    }
                }
            };
        },
        async main(_output, info) {
            const introspections = await info.helpers.pgIntrospection.getIntrospection();
            await Promise.all(introspections.map(async (result) => {
                const { introspection, pgConfig } = result;
                const { namespaces, classes, attributes, constraints, procs, roles, auth_members, types, enums, extensions, indexes, languages, ranges, depends, descriptions, } = introspection;
                function announce(eventName, entities) {
                    const promises = [];
                    for (const entity of entities) {
                        promises.push(info.process(eventName, {
                            entity: entity,
                            databaseName: pgConfig.name,
                        }));
                    }
                    return Promise.all(promises);
                }
                await info.process("pgIntrospection_introspection", {
                    introspection,
                    databaseName: pgConfig.name,
                });
                await announce("pgIntrospection_namespace", namespaces);
                await announce("pgIntrospection_class", classes);
                await announce("pgIntrospection_attribute", attributes);
                await announce("pgIntrospection_constraint", constraints);
                await announce("pgIntrospection_proc", procs);
                await announce("pgIntrospection_role", roles);
                await announce("pgIntrospection_auth_member", auth_members);
                await announce("pgIntrospection_type", types);
                await announce("pgIntrospection_enum", enums);
                await announce("pgIntrospection_extension", extensions);
                await announce("pgIntrospection_index", indexes);
                await announce("pgIntrospection_language", languages);
                await announce("pgIntrospection_range", ranges);
                await announce("pgIntrospection_depend", depends);
                await announce("pgIntrospection_description", descriptions);
            }));
        },
    },
};
//# sourceMappingURL=PgIntrospectionPlugin.js.map