"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PgRelationsPlugin = void 0;
const tslib_1 = require("tslib");
require("graphile-build");
require("./PgTablesPlugin.js");
require("graphile-config");
const pg_1 = require("@dataplan/pg");
const grafast_1 = require("grafast");
const graphile_export_1 = require("graphile-export");
const pg_sql2_1 = tslib_1.__importDefault(require("pg-sql2"));
const tamedevil_1 = tslib_1.__importStar(require("tamedevil"));
const behavior_js_1 = require("../behavior.js");
const utils_js_1 = require("../utils.js");
const version_js_1 = require("../version.js");
// TODO: split this into one plugin for gathering and another for schema
exports.PgRelationsPlugin = {
    name: "PgRelationsPlugin",
    description: "Creates relationships between the @dataplan/pg data sources, and mirrors these relationships into the GraphQL schema",
    version: version_js_1.version,
    after: ["smart-tags", "PgFakeConstraintsPlugin", "PgTablesPlugin"],
    inflection: {
        add: {
            sourceRelationName(options, { databaseName, pgConstraint, localClass: _localClass, localColumns, foreignClass, foreignColumns, isUnique, isReferencee, }) {
                const { tags } = pgConstraint.getTagsAndDescription();
                if (!isReferencee && typeof tags.fieldName === "string") {
                    return tags.fieldName;
                }
                if (isReferencee && typeof tags.foreignFieldName === "string") {
                    return tags.foreignFieldName;
                }
                const remoteName = this.tableSourceName({
                    databaseName,
                    pgClass: foreignClass,
                });
                const columns = !isReferencee
                    ? // We have a column referencing another table
                        localColumns
                    : // The other table has a constraint that references us; this is the backwards relation.
                        foreignColumns;
                const columnNames = columns.map((col) => col.attname);
                return this.camelCase(`${isUnique ? remoteName : this.pluralize(remoteName)}-by-${isReferencee ? "their" : "my"}-${columnNames.join("-and-")}`);
            },
            singleRelation(options, details) {
                const { source, relationName } = details;
                const relation = source.getRelation(relationName);
                //const codec = relation.source.codec;
                if (typeof relation.extensions?.tags.fieldName === "string") {
                    return relation.extensions.tags.fieldName;
                }
                // E.g. posts(author_id) references users(id)
                const remoteType = this.tableType(relation.source.codec);
                const localColumns = relation.localColumns;
                return this.camelCase(`${remoteType}-by-${this._joinColumnNames(source.codec, localColumns)}`);
            },
            singleRelationBackwards(options, details) {
                const { source, relationName } = details;
                const relation = source.getRelation(relationName);
                if (typeof relation.extensions?.tags.foreignSingleFieldName === "string") {
                    return relation.extensions.tags.foreignSingleFieldName;
                }
                if (typeof relation.extensions?.tags.foreignFieldName === "string") {
                    return relation.extensions.tags.foreignFieldName;
                }
                // E.g. posts(author_id) references users(id)
                const remoteType = this.tableType(relation.source.codec);
                const remoteColumns = relation.remoteColumns;
                return this.camelCase(`${remoteType}-by-${this._joinColumnNames(relation.source.codec, remoteColumns)}`);
            },
            _manyRelation(options, details) {
                const { source, relationName } = details;
                const relation = source.getRelation(relationName);
                const baseOverride = relation.extensions?.tags.foreignFieldName;
                if (typeof baseOverride === "string") {
                    return baseOverride;
                }
                // E.g. users(id) references posts(author_id)
                const remoteType = this.tableType(relation.source.codec);
                const remoteColumns = relation.remoteColumns;
                return this.camelCase(`${this.pluralize(remoteType)}-by-${this._joinColumnNames(relation.source.codec, remoteColumns)}`);
            },
            manyRelationConnection(options, details) {
                const { source, relationName } = details;
                const relation = source.getRelation(relationName);
                const override = relation.extensions?.tags.foreignConnectionFieldName;
                if (typeof override === "string") {
                    return override;
                }
                return this.connectionField(this._manyRelation(details));
            },
            manyRelationList(options, details) {
                const { source, relationName } = details;
                const relation = source.getRelation(relationName);
                const override = relation.extensions?.tags.foreignSimpleFieldName;
                if (typeof override === "string") {
                    return override;
                }
                return this.listField(this._manyRelation(details));
            },
        },
    },
    gather: {
        namespace: "pgRelations",
        initialState: () => Object.create(null),
        helpers: {
            async addRelation(info, event, pgConstraint, isReferencee = false) {
                const pgClass = isReferencee
                    ? pgConstraint.getForeignClass()
                    : pgConstraint.getClass();
                const foreignClass = isReferencee
                    ? pgConstraint.getClass()
                    : pgConstraint.getForeignClass();
                if (!pgClass || !foreignClass) {
                    throw new Error(`Invalid introspection`);
                }
                const localColumnNumbers = isReferencee
                    ? pgConstraint.confkey
                    : pgConstraint.conkey;
                const foreignColumnNumbers = isReferencee
                    ? pgConstraint.conkey
                    : pgConstraint.confkey;
                const isUnique = !isReferencee
                    ? true
                    : (() => {
                        // This relationship is unique if the REFERENCED table (not us!)
                        // has a unique constraint on the remoteColumns the relationship
                        // specifies (or a subset thereof).
                        const foreignUniqueColumnOnlyConstraints = foreignClass
                            .getConstraints()
                            .filter((c) => ["u", "p"].includes(c.contype) &&
                            c.conkey?.every((k) => k > 0));
                        const foreignUniqueColumnNumberCombinations = foreignUniqueColumnOnlyConstraints.map((c) => c.conkey);
                        const isUnique = foreignUniqueColumnNumberCombinations.some((foreignUniqueColumnNumbers) => {
                            return foreignUniqueColumnNumbers.every((n) => n > 0 && pgConstraint.conkey.includes(n));
                        });
                        return isUnique;
                    })();
                const { databaseName, relations } = event;
                const localColumns = await Promise.all(localColumnNumbers.map((key) => info.helpers.pgIntrospection.getAttribute(databaseName, pgClass._id, key)));
                const foreignColumns = await Promise.all(foreignColumnNumbers.map((key) => info.helpers.pgIntrospection.getAttribute(databaseName, foreignClass._id, key)));
                const foreignSource = await info.helpers.pgTables.getSourceBuilder(databaseName, foreignClass);
                if (!foreignSource || foreignSource.isVirtual) {
                    return;
                }
                const relationName = info.inflection.sourceRelationName({
                    databaseName,
                    pgConstraint,
                    localClass: pgClass,
                    localColumns: localColumns,
                    foreignClass,
                    foreignColumns: foreignColumns,
                    isUnique,
                    isReferencee,
                });
                const existingRelation = relations[relationName];
                const { tags: rawTags, description: constraintDescription } = pgConstraint.getTagsAndDescription();
                // Clone the tags because we use the same tags on both relations
                // (in both directions) but don't want modifications made to one
                // to affect the other.
                const tags = JSON.parse(JSON.stringify(rawTags));
                const description = isReferencee
                    ? tags.backwardDescription
                    : tags.forwardDescription ?? constraintDescription;
                const newRelation = {
                    localColumns: localColumns.map((c) => c.attname),
                    remoteColumns: foreignColumns.map((c) => c.attname),
                    source: foreignSource,
                    isUnique,
                    isReferencee,
                    description: typeof description === "string" ? description : undefined,
                    extensions: {
                        tags,
                    },
                };
                await info.process("pgRelations_relation", {
                    databaseName,
                    pgClass,
                    pgConstraint,
                    relation: newRelation,
                });
                if (existingRelation) {
                    const isEquivalent = existingRelation.isUnique === newRelation.isUnique &&
                        existingRelation.isReferencee === newRelation.isReferencee &&
                        (0, grafast_1.arraysMatch)(existingRelation.localColumns, newRelation.localColumns) &&
                        (0, grafast_1.arraysMatch)(existingRelation.remoteColumns, newRelation.remoteColumns) &&
                        existingRelation.source === newRelation.source;
                    const message = `Attempted to add a ${isReferencee ? "backwards" : "forwards"} relation named '${relationName}' to '${pgClass.relname}' for ${isEquivalent ? "equivalent " : ""}constraint '${pgConstraint.conname}' on '${pgClass.getNamespace().nspname}.${pgClass.relname}', but a relation by that name already exists; consider renaming the relation by overriding the 'sourceRelationName' inflector`;
                    if (isEquivalent) {
                        console.warn(message);
                        return;
                    }
                    else {
                        throw new Error(message);
                    }
                }
                relations[relationName] = newRelation;
            },
        },
        hooks: {
            async pgTables_PgSourceBuilder_relations(info, event) {
                const { pgClass, databaseName } = event;
                const constraints = await info.helpers.pgIntrospection.getConstraintsForClass(databaseName, pgClass._id);
                const foreignConstraints = await info.helpers.pgIntrospection.getForeignConstraintsForClass(databaseName, pgClass._id);
                for (const constraint of constraints) {
                    if (constraint.contype === "f") {
                        await info.helpers.pgRelations.addRelation(event, constraint);
                    }
                }
                for (const constraint of foreignConstraints) {
                    if (constraint.contype === "f") {
                        await info.helpers.pgRelations.addRelation(event, constraint, true);
                    }
                }
            },
        },
    },
    schema: {
        hooks: {
            GraphQLInterfaceType_fields: addRelations,
            GraphQLObjectType_fields: addRelations,
        },
    },
};
function makeSpecString(identifier, localColumns, remoteColumns) {
    return (0, tamedevil_1.default) `{ ${tamedevil_1.default.join(remoteColumns.map((remoteColumnName, i) => (0, tamedevil_1.default) `${tamedevil_1.default.dangerousKey(remoteColumnName)}: ${identifier}.get(${tamedevil_1.default.lit(localColumns[i])})`), ", ")} }`;
}
function makeRelationPlans(localColumns, remoteColumns, otherSource, isMutationPayload) {
    const recordOrResult = isMutationPayload
        ? (0, tamedevil_1.default) `$record.get("result")`
        : (0, tamedevil_1.default) `$record`;
    const clean = remoteColumns.every((remoteColumnName) => typeof remoteColumnName === "string" &&
        (0, tamedevil_1.isSafeObjectPropertyName)(remoteColumnName)) &&
        localColumns.every((localColumnName) => typeof localColumnName === "string" &&
            (0, tamedevil_1.isSafeObjectPropertyName)(localColumnName));
    const specString = clean
        ? makeSpecString(recordOrResult, localColumns, remoteColumns)
        : null;
    const specFromRecord = (0, graphile_export_1.EXPORTABLE)((localColumns, remoteColumns) => ($record) => {
        return remoteColumns.reduce((memo, remoteColumnName, i) => {
            memo[remoteColumnName] = $record.get(localColumns[i]);
            return memo;
        }, Object.create(null));
    }, [localColumns, remoteColumns]);
    const singleRecordPlan = clean && specString
        ? // Optimise function for both execution and export.
            // eslint-disable-next-line graphile-export/exhaustive-deps
            (0, graphile_export_1.EXPORTABLE)(tamedevil_1.default.run `\
return function (otherSource) {
  return $record => otherSource.get(${specString});
}`, [otherSource])
        : isMutationPayload
            ? (0, graphile_export_1.EXPORTABLE)((otherSource, specFromRecord) => function plan($in) {
                const $record = $in.get("result");
                return otherSource.get(specFromRecord($record));
            }, [otherSource, specFromRecord])
            : (0, graphile_export_1.EXPORTABLE)((otherSource, specFromRecord) => function plan($record) {
                return otherSource.get(specFromRecord($record));
            }, [otherSource, specFromRecord]);
    const listPlan = clean && specString
        ? // eslint-disable-next-line graphile-export/exhaustive-deps
            (0, graphile_export_1.EXPORTABLE)(tamedevil_1.default.run `\
return function (otherSource) {
  return $record => otherSource.find(${specString});
}`, [otherSource])
        : isMutationPayload
            ? (0, graphile_export_1.EXPORTABLE)((otherSource, specFromRecord) => function plan($in) {
                const $record = $in.get("result");
                return otherSource.find(specFromRecord($record));
            }, [otherSource, specFromRecord])
            : (0, graphile_export_1.EXPORTABLE)((otherSource, specFromRecord) => function plan($record) {
                return otherSource.find(specFromRecord($record));
            }, [otherSource, specFromRecord]);
    const connectionPlan = clean && specString
        ? // eslint-disable-next-line graphile-export/exhaustive-deps
            (0, graphile_export_1.EXPORTABLE)(tamedevil_1.default.run `\
return function (otherSource, connection) {
  return $record => {
    const $records = otherSource.find(${specString});
    return connection($records);
  }
}`, [otherSource, grafast_1.connection])
        : isMutationPayload
            ? (0, graphile_export_1.EXPORTABLE)((connection, otherSource, specFromRecord) => function plan($in) {
                const $record = $in.get("result");
                return connection(otherSource.find(specFromRecord($record)));
            }, [grafast_1.connection, otherSource, specFromRecord])
            : (0, graphile_export_1.EXPORTABLE)((connection, otherSource, specFromRecord) => function plan($record) {
                return connection(otherSource.find(specFromRecord($record)));
            }, [grafast_1.connection, otherSource, specFromRecord]);
    return { singleRecordPlan, listPlan, connectionPlan };
}
function addRelations(fields, build, context) {
    const { extend, graphql: { GraphQLList, GraphQLObjectType, GraphQLNonNull, GraphQLUnionType, GraphQLInterfaceType, }, options: { simpleCollections }, } = build;
    const { Self, scope, fieldWithHooks } = context;
    // const objectMode = context.type === "GraphQLObjectType";
    const { pgCodec } = scope;
    const isPgTableType = "isPgTableType" in scope ? scope.isPgTableType : undefined;
    const pgTypeSource = "pgTypeSource" in scope ? scope.pgTypeSource : undefined;
    const isMutationPayload = "isMutationPayload" in scope ? scope.isMutationPayload : undefined;
    const pgPolymorphism = "pgPolymorphism" in scope ? scope.pgPolymorphism : undefined;
    const codec = pgTypeSource?.codec ?? pgCodec;
    // TODO: make it so isMutationPayload doesn't trigger this by default (only in V4 compatibility mode)
    if (!(isPgTableType || isMutationPayload || pgPolymorphism) || !codec) {
        return fields;
    }
    // TODO: change the default so that we don't do this on
    // isMutationPayload; only do that for V4 compat. (It's redundant vs
    // just using the object type directly)
    const source = pgTypeSource ??
        build.input.pgSources.find((s) => s.codec === codec && !s.parameters) ??
        build.input.pgSources.find((s) => s.codec === codec && s.isUnique);
    if (!source && !codec.extensions?.refDefinitions) {
        return fields;
    }
    if (source && source.parameters && !source.isUnique) {
        return fields;
    }
    const relations = source?.getRelations();
    // Don't use refs on mutation payloads
    const refDefinitionList = isMutationPayload
        ? []
        : source
            ? Object.entries(source.refs).map(([refName, spec]) => ({
                refName,
                refDefinition: spec.definition,
                ref: spec,
            }))
            : Object.entries(codec.extensions?.refDefinitions ??
                Object.create(null)).map(([refName, refDefinition]) => ({
                refName,
                refDefinition,
                codec,
            }));
    const resolvePath = (path) => {
        if (!source) {
            throw new Error(`Cannot call resolvePath unless there's a source`);
        }
        const result = {
            source: source,
            hasReferencee: false,
            isUnique: true,
            layers: [],
        };
        for (const pathEntry of path) {
            const relation = result.source.getRelation(pathEntry.relationName);
            const { isReferencee, localColumns, remoteColumns, source: sourceOrBuilder, isUnique, } = relation;
            const source = sourceOrBuilder instanceof pg_1.PgSourceBuilder
                ? sourceOrBuilder.get()
                : sourceOrBuilder;
            if (isReferencee) {
                result.hasReferencee = true;
            }
            if (!isUnique) {
                result.isUnique = false;
            }
            result.layers.push({
                relationName: pathEntry.relationName,
                localColumns: localColumns,
                remoteColumns: remoteColumns,
                source,
                isUnique,
            });
            result.source =
                relation.source instanceof pg_1.PgSourceBuilder
                    ? relation.source.get()
                    : relation.source;
        }
        return result;
    };
    const digests = [];
    if (source) {
        // Digest relations
        for (const [relationName, relation] of Object.entries(relations)) {
            const { localColumns, remoteColumns, source: otherSourceOrBuilder, extensions, isReferencee, } = relation;
            if (isMutationPayload && isReferencee) {
                // Don't add backwards relations to mutation payloads
                continue;
            }
            const otherSource = otherSourceOrBuilder instanceof pg_1.PgSourceBuilder
                ? otherSourceOrBuilder.get()
                : otherSourceOrBuilder;
            // The behavior is the relation behavior PLUS the remote table
            // behavior. But the relation settings win.
            const behavior = (0, behavior_js_1.getBehavior)([
                otherSource.codec.extensions,
                otherSource.extensions,
                extensions,
            ]) ?? "";
            const otherCodec = otherSource.codec;
            const typeName = build.inflection.tableType(otherCodec);
            const connectionTypeName = build.inflection.tableConnectionType(otherCodec);
            const deprecationReason = (0, utils_js_1.tagToString)(relation.extensions?.tags?.deprecated) ??
                (0, utils_js_1.tagToString)(relation.source.extensions?.tags?.deprecated);
            const relationDetails = {
                source,
                relationName,
            };
            const { singleRecordPlan, listPlan, connectionPlan } = makeRelationPlans(localColumns, remoteColumns, otherSource, isMutationPayload ?? false);
            const singleRecordFieldName = relation.isReferencee
                ? build.inflection.singleRelationBackwards(relationDetails)
                : build.inflection.singleRelation(relationDetails);
            const connectionFieldName = build.inflection.manyRelationConnection(relationDetails);
            const listFieldName = build.inflection.manyRelationList(relationDetails);
            const digest = {
                identifier: relationName,
                isReferencee: relation.isReferencee ?? false,
                behavior,
                isUnique: relation.isUnique,
                typeName,
                connectionTypeName,
                deprecationReason,
                singleRecordPlan,
                listPlan,
                connectionPlan,
                singleRecordFieldName,
                listFieldName,
                connectionFieldName,
                description: relation.description,
                pgSource: otherSource,
                pgCodec: otherSource.codec,
                pgRelationDetails: relationDetails,
                relatedTypeName: build.inflection.tableType(codec),
            };
            digests.push(digest);
        }
    }
    // Digest refs
    for (const { refName: identifier, refDefinition: refSpec, ref, } of refDefinitionList) {
        let hasReferencee;
        let sharedCodec = undefined;
        let sharedSource = undefined;
        let behavior;
        let typeName;
        let singleRecordPlan;
        let listPlan;
        let connectionPlan;
        if (ref && source) {
            const paths = ref.paths.map(resolvePath);
            if (paths.length === 0)
                continue;
            const firstSource = paths[0].source;
            const hasExactlyOneSource = paths.every((p) => p.source === firstSource);
            const firstCodec = firstSource.codec;
            const hasExactlyOneCodec = paths.every((p) => p.source.codec === firstCodec);
            hasReferencee = paths.some((p) => p.hasReferencee);
            if (isMutationPayload && (paths.length !== 1 || hasReferencee)) {
                // Don't add backwards relations to mutation payloads
                continue;
            }
            typeName =
                refSpec.graphqlType ??
                    (hasExactlyOneCodec ? build.inflection.tableType(firstCodec) : null);
            if (!typeName) {
                continue;
            }
            const type = build.getTypeByName(typeName);
            if (!type) {
                continue;
            }
            if (refSpec.graphqlType) {
                // If this is a union/interface, can we find the associated codec?
                const scope = build.scopeByType.get(type);
                if (scope) {
                    if ("pgCodec" in scope) {
                        sharedCodec = scope.pgCodec;
                    }
                }
            }
            else if (hasExactlyOneCodec) {
                sharedCodec = firstCodec;
            }
            if (hasExactlyOneSource) {
                sharedSource = firstSource;
            }
            // TODO: if there's only one path do we still need union?
            const needsPgUnionAll = sharedCodec?.polymorphism?.mode === "union" || paths.length > 1;
            // If we're pulling from a shared codec into a PgUnionAllStep then we can
            // use that codec's columns as shared attributes; otherwise there are not
            // shared attributes (equivalent to a GraphQL union).
            const unionAttributes = sharedCodec?.columns;
            // const isUnique = paths.every((p) => p.isUnique);
            // TODO: shouldn't the ref behavior override the source behavior?
            behavior = hasExactlyOneSource
                ? (0, behavior_js_1.getBehavior)([
                    firstSource.codec.extensions,
                    firstSource.extensions,
                    refSpec.extensions,
                ])
                : (0, behavior_js_1.getBehavior)([sharedCodec?.extensions, refSpec.extensions]);
            // Shortcut simple relation alias
            ({ singleRecordPlan, listPlan, connectionPlan } = (() => {
                // Add forbidden names here
                if (ref.paths.length === 1 && ref.paths[0].length === 1) {
                    const relation = source.getRelation(ref.paths[0][0].relationName);
                    const otherSource = relation.source instanceof pg_1.PgSourceBuilder
                        ? relation.source.get()
                        : relation.source;
                    return makeRelationPlans(relation.localColumns, relation.remoteColumns, otherSource, isMutationPayload ?? false);
                }
                else if (!needsPgUnionAll) {
                    // Definitely just one chain
                    const path = paths[0];
                    const makePlanResolver = (mode) => {
                        const single = mode === "singleRecord";
                        const isConnection = mode === "connection";
                        const idents = new tamedevil_1.Idents();
                        idents.forbid([
                            "$list",
                            "$in",
                            "$record",
                            "$entry",
                            "$tuple",
                            "list",
                            "sql",
                        ]);
                        const functionLines = [];
                        const prefixLines = [];
                        if (isMutationPayload) {
                            functionLines.push((0, tamedevil_1.default) `return ($in) => {`);
                            functionLines.push((0, tamedevil_1.default) `  const $record = $in.get("result");`);
                        }
                        else {
                            functionLines.push((0, tamedevil_1.default) `return ($record) => {`);
                        }
                        let previousIdentifier = tamedevil_1.default.identifier(`$record`);
                        // ENHANCEMENT: these ensure that the variables are defined in
                        // the closure, but they also output noise
                        // (`list;object;connection;sql;`) which could be eliminated.
                        prefixLines.push((0, tamedevil_1.default) `${tamedevil_1.default.ref(grafast_1.list, "list")};`);
                        prefixLines.push((0, tamedevil_1.default) `${tamedevil_1.default.ref(grafast_1.object, "object")};`);
                        prefixLines.push((0, tamedevil_1.default) `${tamedevil_1.default.ref(grafast_1.connection, "connection")};`);
                        prefixLines.push((0, tamedevil_1.default) `${tamedevil_1.default.ref(pg_sql2_1.default, "sql")};`);
                        let isStillSingular = true;
                        for (let i = 0, l = path.layers.length; i < l; i++) {
                            const layer = path.layers[i];
                            const { localColumns, remoteColumns, source, isUnique } = layer;
                            const clean = localColumns.every(tamedevil_1.isSafeObjectPropertyName) &&
                                remoteColumns.every(tamedevil_1.isSafeObjectPropertyName);
                            const sourceName = idents.makeSafeIdentifier(`${source.name}Source`);
                            prefixLines.push((0, tamedevil_1.default) `const ${tamedevil_1.default.identifier(sourceName)} = ${tamedevil_1.default.ref(source)};`);
                            if (isStillSingular) {
                                if (!isUnique) {
                                    isStillSingular = false;
                                }
                                const newIdentifier = tamedevil_1.default.identifier(idents.makeSafeIdentifier(`$${isUnique
                                    ? build.inflection.singularize(source.name)
                                    : build.inflection.pluralize(source.name)}`));
                                const specFromRecord = (0, graphile_export_1.EXPORTABLE)((localColumns, remoteColumns) => ($record) => {
                                    return remoteColumns.reduce((memo, remoteColumnName, i) => {
                                        memo[remoteColumnName] = $record.get(localColumns[i]);
                                        return memo;
                                    }, Object.create(null));
                                }, [localColumns, remoteColumns]);
                                const specString = clean
                                    ? makeSpecString(previousIdentifier, localColumns, remoteColumns)
                                    : (0, tamedevil_1.default) `${tamedevil_1.default.ref(specFromRecord)}(${previousIdentifier})`;
                                functionLines.push((0, tamedevil_1.default) `  const ${newIdentifier} = ${tamedevil_1.default.identifier(sourceName)}.${isUnique ? (0, tamedevil_1.default) `get` : (0, tamedevil_1.default) `find`}(${specString});`);
                                previousIdentifier = newIdentifier;
                            }
                            else {
                                const newIdentifier = tamedevil_1.default.identifier(idents.makeSafeIdentifier(`$${build.inflection.pluralize(source.name)}`));
                                /*
                            const tupleIdentifier = makeSafeIdentifier(
                              `${previousIdentifier}Tuples`,
                            );
                            functionLines.push(
                              `  const ${tupleIdentifier} = each(${previousIdentifier}, ($entry) => object({ ${localColumns
                                .map(
                                  (c) =>
                                    `${evalSafeProperty(c)}: $entry.get(${JSON.stringify(c)})`,
                                )
                                .join(", ")} }));`,
                            );
                            functionLines.push(`  ${newIdentifier}.where(sql\`\${}\`);`);
                            */
                                // ENHANCEMENT: we could rename this to be the singular of the source name or something
                                const $entry = (0, tamedevil_1.default) `$entry`;
                                const specString = makeSpecString($entry, localColumns, remoteColumns);
                                functionLines.push((0, tamedevil_1.default) `  const ${newIdentifier} = each(${previousIdentifier}, (${$entry}) => ${tamedevil_1.default.identifier(sourceName)}.get(${specString}));`);
                                previousIdentifier = newIdentifier;
                            }
                        }
                        if (isStillSingular && !single) {
                            const newIdentifier = tamedevil_1.default.identifier("$list");
                            functionLines.push((0, tamedevil_1.default) `  const ${newIdentifier} = list([${previousIdentifier}]);`);
                            previousIdentifier = newIdentifier;
                        }
                        if (isConnection) {
                            functionLines.push((0, tamedevil_1.default) `  return connection(${previousIdentifier});`);
                        }
                        else {
                            functionLines.push((0, tamedevil_1.default) `  return ${previousIdentifier};`);
                        }
                        functionLines.push((0, tamedevil_1.default) `}`);
                        return tamedevil_1.default.run `${tamedevil_1.default.join(prefixLines, "\n")}${tamedevil_1.default.join(functionLines, "\n")}`;
                    };
                    const singleRecordPlan = makePlanResolver("singleRecord");
                    const listPlan = makePlanResolver("list");
                    const connectionPlan = makePlanResolver("connection");
                    return { singleRecordPlan, listPlan, connectionPlan };
                }
                else {
                    const makePlanResolver = (mode) => {
                        const single = mode === "singleRecord";
                        const isConnection = mode === "connection";
                        const attributes = unionAttributes ?? {};
                        const sourceByTypeName = Object.create(null);
                        const members = [];
                        for (const path of paths) {
                            const [firstLayer, ...rest] = path.layers;
                            const memberPath = [];
                            let finalSource = firstLayer.source;
                            for (const layer of rest) {
                                const { relationName } = layer;
                                memberPath.push({ relationName });
                                finalSource = layer.source;
                            }
                            const typeName = build.inflection.tableType(finalSource.codec);
                            const member = {
                                source: firstLayer.source,
                                typeName,
                                path: memberPath,
                            };
                            members.push(member);
                            if (!sourceByTypeName[typeName]) {
                                sourceByTypeName[typeName] = finalSource;
                            }
                        }
                        return (0, graphile_export_1.EXPORTABLE)((attributes, connection, isConnection, isMutationPayload, members, paths, pgUnionAll, single, sourceByTypeName) => ($parent) => {
                            const $record = isMutationPayload
                                ? $parent.get("result")
                                : $parent;
                            for (let i = 0, l = paths.length; i < l; i++) {
                                const path = paths[i];
                                const firstLayer = path.layers[0];
                                const member = members[i];
                                member.match = firstLayer.localColumns.reduce((memo, col, idx) => {
                                    memo[firstLayer.remoteColumns[idx]] = {
                                        step: $record.get(col),
                                    };
                                    return memo;
                                }, Object.create(null));
                            }
                            const $list = pgUnionAll({
                                attributes,
                                sourceByTypeName,
                                members,
                            });
                            if (isConnection) {
                                return connection($list);
                            }
                            else if (single) {
                                return $list.single();
                            }
                            else {
                                return $list;
                            }
                        }, [
                            attributes,
                            grafast_1.connection,
                            isConnection,
                            isMutationPayload,
                            members,
                            paths,
                            pg_1.pgUnionAll,
                            single,
                            sourceByTypeName,
                        ]);
                    };
                    const singleRecordPlan = makePlanResolver("singleRecord");
                    const listPlan = makePlanResolver("list");
                    const connectionPlan = makePlanResolver("connection");
                    return {
                        singleRecordPlan,
                        listPlan,
                        connectionPlan,
                    };
                }
            })());
        }
        else {
            hasReferencee = true;
            behavior = (0, behavior_js_1.getBehavior)(refSpec.extensions);
            typeName = refSpec.graphqlType;
            if (!typeName) {
                // TODO: remove this restriction
                throw new Error(`@ref on polymorphic type must declare to:TargetType`);
            }
            const type = build.getTypeByName(typeName);
            if (!type) {
                continue;
            }
            if (refSpec.graphqlType) {
                // If this is a union/interface, can we find the associated codec?
                const scope = build.scopeByType.get(type);
                if (scope) {
                    if ("pgCodec" in scope) {
                        sharedCodec = scope.pgCodec;
                    }
                }
            }
        }
        const connectionTypeName = sharedCodec
            ? build.inflection.tableConnectionType(sharedCodec)
            : build.inflection.connectionType(typeName);
        const singleRecordFieldName = build.inflection.refSingle({
            refDefinition: refSpec,
            identifier,
        });
        const connectionFieldName = build.inflection.refConnection({
            refDefinition: refSpec,
            identifier,
        });
        const listFieldName = build.inflection.refList({
            refDefinition: refSpec,
            identifier,
        });
        const digest = {
            identifier,
            isReferencee: hasReferencee,
            pgCodec: sharedCodec,
            pgSource: sharedSource,
            isUnique: !!refSpec.singular,
            behavior: behavior ?? "",
            typeName,
            connectionTypeName,
            singleRecordFieldName,
            connectionFieldName,
            listFieldName,
            singleRecordPlan,
            listPlan,
            connectionPlan,
            relatedTypeName: context.Self.name,
        };
        digests.push(digest);
    }
    return digests.reduce((memo, digest) => {
        const { isUnique, behavior, typeName, connectionTypeName, deprecationReason, singleRecordFieldName, listFieldName, connectionFieldName, singleRecordPlan, listPlan, connectionPlan, isReferencee, identifier, description, pgSource, pgCodec: pgFieldCodec, pgRelationDetails, relatedTypeName, } = digest;
        const relationTypeScope = isUnique ? `singularRelation` : `manyRelation`;
        const OtherType = build.getTypeByName(typeName);
        if (!OtherType ||
            !(OtherType instanceof GraphQLObjectType ||
                OtherType instanceof GraphQLInterfaceType ||
                OtherType instanceof GraphQLUnionType)) {
            return memo;
        }
        let fields = memo;
        const defaultBehavior = isUnique
            ? "single -singularRelation:source:list -singularRelation:source:connection"
            : simpleCollections === "both"
                ? "connection list"
                : simpleCollections === "only"
                    ? "list"
                    : "connection";
        if (isUnique &&
            build.behavior.matches(behavior, `${relationTypeScope}:source:single`, defaultBehavior)) {
            const fieldName = singleRecordFieldName;
            fields = extend(fields, {
                [fieldName]: fieldWithHooks({
                    fieldName,
                    fieldBehaviorScope: `${relationTypeScope}:source:single`,
                    isPgSingleRelationField: true,
                    behavior,
                    pgRelationDetails,
                }, {
                    description: description ??
                        `Reads a single \`${typeName}\` that is related to this \`${relatedTypeName}\`.`,
                    // TODO: handle nullability
                    type: OtherType,
                    plan: singleRecordPlan,
                    deprecationReason,
                }),
            }, `Adding '${identifier}' single relation field to ${Self.name}`, "recoverable");
        }
        if (isReferencee &&
            build.behavior.matches(behavior, `${relationTypeScope}:source:connection`, defaultBehavior)) {
            const ConnectionType = build.getTypeByName(connectionTypeName);
            if (ConnectionType) {
                const fieldName = connectionFieldName;
                fields = extend(fields, {
                    [fieldName]: fieldWithHooks({
                        fieldName,
                        fieldBehaviorScope: `${relationTypeScope}:source:connection`,
                        // TODO: rename to pgFieldSource?
                        pgSource,
                        pgFieldCodec,
                        isPgFieldConnection: true,
                        isPgManyRelationConnectionField: true,
                        // TODO: rename to pgFieldRelationDetails?
                        pgRelationDetails,
                        // TODO: rename to pgFieldBehavior?
                        behavior,
                    }, {
                        description: description ??
                            `Reads and enables pagination through a set of \`${typeName}\`.`,
                        // TODO: handle nullability
                        type: new GraphQLNonNull(ConnectionType),
                        plan: connectionPlan,
                        deprecationReason,
                    }),
                }, `Adding '${identifier}' many relation connection field to ${Self.name}`, "recoverable");
            }
            else {
                console.log(`Could not find connection type '${connectionTypeName}'`);
            }
        }
        if (isReferencee &&
            build.behavior.matches(behavior, `${relationTypeScope}:source:list`, defaultBehavior)) {
            const fieldName = listFieldName;
            fields = extend(fields, {
                [fieldName]: fieldWithHooks({
                    fieldName,
                    fieldBehaviorScope: `${relationTypeScope}:source:list`,
                    pgSource,
                    pgFieldCodec,
                    isPgFieldSimpleCollection: true,
                    isPgManyRelationListField: true,
                    pgRelationDetails,
                    behavior,
                }, {
                    description: description ??
                        `Reads and enables pagination through a set of \`${typeName}\`.`,
                    type: new GraphQLNonNull(new GraphQLList(new GraphQLNonNull(OtherType))),
                    plan: listPlan,
                    deprecationReason,
                }),
            }, `Adding '${identifier}' many relation list field to ${Self.name}`, "recoverable");
        }
        return fields;
    }, fields);
}
//# sourceMappingURL=PgRelationsPlugin.js.map