"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PgTableNodePlugin = void 0;
const tslib_1 = require("tslib");
require("graphile-build");
require("graphile-config");
const grafast_1 = require("grafast");
const graphile_export_1 = require("graphile-export");
const tamedevil_1 = tslib_1.__importStar(require("tamedevil"));
const behavior_js_1 = require("../behavior.js");
const utils_js_1 = require("../utils.js");
const version_js_1 = require("../version.js");
exports.PgTableNodePlugin = {
    name: "PgTableNodePlugin",
    description: "Add the 'Node' interface to table types",
    version: version_js_1.version,
    schema: {
        hooks: {
            init(_, build) {
                if (!build.registerNodeIdHandler) {
                    return _;
                }
                const tableSources = build.input.pgSources.filter((source) => {
                    if (source.codec.isAnonymous)
                        return false;
                    if (!source.codec.columns)
                        return false;
                    if (source.codec.polymorphism)
                        return false;
                    if (source.parameters)
                        return false;
                    if (!source.uniques)
                        return false;
                    if (!source.uniques[0])
                        return false;
                    const behavior = (0, behavior_js_1.getBehavior)([
                        source.codec.extensions,
                        source.extensions,
                    ]);
                    // Needs the 'select' and 'node' behaviours for compatibility
                    return (!!build.behavior.matches(behavior, "node", "node") &&
                        !!build.behavior.matches(behavior, "select", "select"));
                });
                const sourcesByCodec = new Map();
                for (const source of tableSources) {
                    let list = sourcesByCodec.get(source.codec);
                    if (!list) {
                        list = [];
                        sourcesByCodec.set(source.codec, list);
                    }
                    list.push(source);
                }
                for (const [codec, sources] of sourcesByCodec.entries()) {
                    const tableTypeName = build.inflection.tableType(codec);
                    if (sources.length !== 1) {
                        console.warn(`Found multiple table sources for codec '${codec.name}'; we don't currently support that but we _could_ - get in touch if you need this.`);
                        continue;
                    }
                    const pgSource = sources[0];
                    const primaryKey = pgSource.uniques.find((u) => u.isPrimary === true);
                    if (!primaryKey) {
                        continue;
                    }
                    const pk = primaryKey.columns;
                    const identifier = 
                    // Yes, this behaviour in V4 was ridiculous. Alas.
                    build.options.pgV4UseTableNameForNodeIdentifier &&
                        pgSource.extensions?.pg?.name
                        ? build.inflection.pluralize(pgSource.extensions.pg.name)
                        : tableTypeName;
                    const clean = (0, tamedevil_1.isSafeObjectPropertyName)(identifier) &&
                        pk.every((columnName) => (0, tamedevil_1.isSafeObjectPropertyName)(columnName));
                    const firstSource = sources.find((s) => !s.parameters);
                    build.registerNodeIdHandler({
                        typeName: tableTypeName,
                        codecName: "base64JSON",
                        deprecationReason: (0, utils_js_1.tagToString)(codec.extensions?.tags?.deprecation ??
                            firstSource?.extensions?.tags?.deprecated),
                        plan: clean
                            ? // eslint-disable-next-line graphile-export/exhaustive-deps
                                (0, graphile_export_1.EXPORTABLE)(tamedevil_1.default.run `\
return function (list, constant) {
  return $record => list([constant(${tamedevil_1.default.lit(identifier)}), ${tamedevil_1.default.join(pk.map((columnName) => (0, tamedevil_1.default) `$record.get(${tamedevil_1.default.lit(columnName)})`), ", ")}]);
}`, [grafast_1.list, grafast_1.constant])
                            : (0, graphile_export_1.EXPORTABLE)((constant, identifier, list, pk) => ($record) => {
                                return list([
                                    constant(identifier),
                                    ...pk.map((column) => $record.get(column)),
                                ]);
                            }, [grafast_1.constant, identifier, grafast_1.list, pk]),
                        getSpec: clean
                            ? // eslint-disable-next-line graphile-export/exhaustive-deps
                                (0, graphile_export_1.EXPORTABLE)(tamedevil_1.default.run `\
return function (access) {
  return $list => ({ ${tamedevil_1.default.join(pk.map((columnName, index) => (0, tamedevil_1.default) `${tamedevil_1.default.dangerousKey(columnName)}: access($list, [${tamedevil_1.default.lit(index + 1)}])`), ", ")} });
}`, [grafast_1.access])
                            : (0, graphile_export_1.EXPORTABLE)((access, pk) => ($list) => {
                                const spec = pk.reduce((memo, column, index) => {
                                    memo[column] = access($list, [index + 1]);
                                    return memo;
                                }, Object.create(null));
                                return spec;
                            }, [grafast_1.access, pk]),
                        get: (0, graphile_export_1.EXPORTABLE)((pgSource) => (spec) => pgSource.get(spec), [pgSource]),
                        match: (0, graphile_export_1.EXPORTABLE)((identifier) => (obj) => {
                            return obj[0] === identifier;
                        }, [identifier]),
                    });
                }
                return _;
            },
        },
    },
};
//# sourceMappingURL=PgTableNodePlugin.js.map